/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

class PatternSelector : Gtk.ScrolledWindow {
    /**
     * Emitted whenever a pattern is selected.
     */
    public signal void pattern_changed_signal (string move);

    /**
     * List of core patterns: randomizers and traditional patterns.
     * This list is fixed: no items are added or removed from it.
     */
    Pattern[] core_patterns = {              
        new RandomPattern (N_("Gardens"),           " J J J O X OfJf"),
        new RandomPattern (N_("Day and Night"),     " d"),
        new RandomPattern (N_("Fans"),              " R R R * *fRf Rf*f"),
        new RandomPattern (N_("Folds"),             " Rt Rt Rt *t RtJf JfRt"),
        new RandomPattern (N_("Oblique Folds"),     " JL JL Xs JLJf JfJL"),
        new RandomPattern (N_("Whatever"),          " RR RR RR *? ** *?*? d"),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Ajong Dengan Sauhnya"),
            "HPHHPPPHHPH PPHPPHPPHPP HHPHHPHHPHH HPHPHPHPHPH PPHHPHPHHPP PHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Awan Berturus"),
            "PHPPHHHPPHP HPHPPPPPHPH PHPHHPHHPHP PPHPHHHPHPP HPHHPHPHHPH HPPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Awan Mengandung Hujan"),
            "PPHPHPPHHHP HPPHPHPPHPP HHHPHPHHHPH HPHHHPPHPHP PPHPPHPHHPH PHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Awan Tunjung"),
            "PHHPPHPPHPP PPHHPHPHPPH HPPHHPHHPHH PHHHHPPHHHP PPHPPHPHHPP HHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bangkawat / Awan Menurus"),
            "PHHPPHHPPHP HHPPHHPPHHH HPPHHPPHPHP PPHHPPHHHPP PHHPPHPHPPH HHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bangkawat Kudung"),
            "PPHHPPHHHHP HPPHHPPHHPP HHPPHHPHPPH HHHPPHHPPHH HPPHPPPPHHP PPHHP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bintang Terhambur"),
            "HPHPHPHPHPH PHPHPHPHPHP HPHPHPHPHPH PHPHPHPHPHP HPHPHPHPHPH PHPHP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bunga Bercucuk"),
            "PPHHHHHHHPP PPPHHPHHPPP HPPPHPHPPPH HHPPPHPPPHH HHHPPHPPHHH HPPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bunga Kambang Kiapu"),
            "HPHHPPPHHPH PPHPHPHPHPP HHPHPHPHPHH HPHPHPHPHPH PHPHPHPHPHP PPHPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bunga Kembang Berhujan"),
            "PPHPPHPPHHP HPHHPHHPHPP HHPPHHPHPHH PPHHPPHHPHP PHPHHPHPHPP HHHPP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Bunga Sepapan"),
            "PPHPPHPPHPP PHHHPHPHHHP HHPPHHHPPHH PHPHHPPHPHP PPHPPHPHHPP HHHPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Galap Madam"),
            "PPPPPHHHHHP HHHHHPPPPHP HPPPPHHHPHP HPHHHPPHPHP HPHPPHPHPHP HPHPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Ikan Bernang"),
            "PHHHPPPHHHP HPPHHPHHPPH HPPPHPHPPPH HHPPHPHPPHH PHHHPHPHHHP PPPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Kalabutan"),
            "HPPHHPHHPPH PPHPHHHPHPP PHPHPHPHPHP HPHPPHPPHPH HHPPHPHPPHH PHHHP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Kalabutan Berdarah Hitam"),
            "HHPPHPHPPHH HPHHPHPHHPH PHPPHPHPPHP PHPPHPHPPHP HPHHPHPHHPH PHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Kalajangking 1"),
            "HPHPHPHPHPH PPHPPHPPHPP HHPHHPHHPHH PPHPPHPPHPP HPHPHHHPHPH PHPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Kalajangking 2"),
            "HPHHPHPHHPH PPHPPHPPHPP HHPHHPHHPHH HPHPPHPPHPH PPHPHPHPHPP HHPHP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Kedudukan Dian"),
            "PHPPHPHPPHP HPHPHPHPHPH PHPHPHPHPHP PPHPHHHPHPP HHPHPHPHPHH PPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Kina Putar"),
            "HPPHHPPHPHH HHPPHHPPHHP PHHPPHHPHPP HPPHPPHHPPH PPHHHPHPPHH PHHPP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Lasung-Lasung"),
            "HPPPPHHHHPH PHHHHPPPHHP HHPPPHHPPHP HPPHHPHHPHP HPHHPPPHPHP HPHPP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Lungun Sipugut"),
            "PPHPHPHPHPP PPHHPHPHHPP HHPHHPHHPHH PHHPPHPPHHP HPHPHPHPHPH PHPHP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Mahligai Bergantung Bertali"),
            "PHPPHPHPPHP HHHPPHPPHHH PHPHPHPHPHP PPHHHPHHHPP HPPHPHPHPPH PHHPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Mayang Piasau"),
            "PHHPPHHHPPP PHPHPPPHPHH PPHHPHPHHPH HHHPHPHPHHP HPPHPHPHPPP HPHPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Pagar Batu"),
            "PHHPPHPPHHP HPPHHPHHPPH HPPHPHPHPPH PHHPHPHPHHP PHPHPHPHPHP HPHPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Paha Ayam"),
            "PPPPPHHHHHP HPPPPHHHHPP HHPPPHHHPPP HHHPPHHPPPP HHHHPHPPPPP HHHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Paita"),
            "HPPHPHPHPPH PPHHHPHHHPP PHHPPHPPHHP HHPPPHPPPHH PHPPHHHPPHP HPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Patah-Patah"),
            "HPHPPHPPHPH PPHPHHHPHPP HHHPHPHPHHH PPPHPHPHPPP PHHPPHPPHHP HHPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Peria Bulat"),
            "PPHHHPHHHPP PPPHHPHHPPP HPPPHHHPPPH HHPPPHPPPHH HHHPPHPPHHH PPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Peria Buruk Hatinya"),
            "HPPHPHPHPPH PHPHHPHHPHP PPHPPHPPHPP HHPHPHPHPHH PHPPHHHPPHP HPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Peria Buruk Sepanggal"),
            "HPPHPPPHHPH PPPHHHHHPPP HPHHHPPHHPP HHHPHPPPHHH PHPPPHPHHHP PHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Peria Sejambangan"),
            "HPPHPHPHPPH PPHHHPHHHPP PHPPHPHPPHP HHPHPHPHPHH PHHPPHPPHHP HPPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Pucuk Rebung"),
            "HHPHPPPHPHH HPPHHPHHPPH PPPPHPHPPPP HHPPHHHPPHH PHHHPHPHHHP PPPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Puncak Mahligai"),
            "PHPHHPHHPHP HPHPPPPPHPH PHHHPPPHHHP HPHHHPHHHPH HPPHHPHHPPH PPPPP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Saputangan Jatuh Ke Laut"),
            "HPPHHPHHPPH PPHHPHPHHPP PHPPHPHPPHP HHPHHPPHPHH HPHPPHPHHPH PHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Sarimbangun"),
            "HPPHPHPHPPH PPHPPHPPHPP PHPHHHHHPHP HPHPPHPPHPH PPHPHHHPHPP HHHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Selidan Putar"),
            "PHHPHPPHHPP PPHHPPHHPPH HPPHPHHPPHH HHPPHHPPHHP PHHPPHPHPPH PPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Si Pugut Menunggu Beringin Buruk Hatinya"),
            "HPHPPHPPHPH PHPHHPHHPHP HPHPHPHPHPH PHPHPHPHPHP PHHPPHPPHHP HPPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Sisir Sepapan"),
            "HPPHHPPHPPH PPHHPPHHHPP PHPHPHHPPHP HHPPHHPPHHH PHHPPHPHPPH PPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Talapok Berantai"),
            "HPHHHPHHHPH PPPHPHPHPPP HPHPHPHPHPH HHPHPPPHPHH HPHPHPHPHPH PHPPP"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Talapok Sepapan"),
            "HPHHHPHHHPH PPPHPPPHPPP HPHPHPHPHPH HHPPPHPPPHH HPHPHHHPHPH PPPHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Tekuyong Naik Pasang"),
            "PHPHPHPHPHP HPHPHPHPHPH PHPHPHPHPHP HPHPPHPPHPH PHPPHHHPPHP HPHHH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Tikus Naik Ke Durong"),
            "HPHHPPPHHPH PPPHPHPHPPP HPPHHPHHPPH HHHPHPHPHHH PPHHPHPHHPP PHPPH"
        ),
        new Pattern (
            // Pattern name. Don't translate.
            N_("Tudung Dulang"),
            "PHPHPHPHPHP HPPHPHPHPPH PPHHPHPHHPP HHHPPHPPHHH PPPPHHHPPPP HHHHH"
        )
    };

    /**
     * Actual list of patterns. This list contains at least core_patterns.
     * Custom patterns can be added/removed through add_patterns.
     */
    Pattern[] patterns;

    /**
     * Most recently used pattern.
     */
    private int mru_pattern_num = 0;

    /**
     * Similar to gtk sensitive property.
     */
    private bool responsive = true;

    /**
     * True when programatically tweaking the pattern buttons.
     */
    private bool ignore_signal = false;

    /**
     * True when a pattern is clicked programatically.
     * A pattern randomizer is always called when its button is programatically clicked,
     * regardless of its activation.
     */
    private bool self_clicked = false;

    /**
     * Buttons depicting opening patterns, including randomizers.
     */
    private Gtk.RadioButton[] pattern_buttons;

    /**
     * Invisible button to support unselect.
     */
    private Gtk.RadioButton invisible_button;

    /**
     * Visual container to hold list of patterns.
     */
    private Gtk.VBox pattern_box;

    /**
     * Used to show a hot randomizer cooling down.
     */
    private Timer timer = null;

    public PatternSelector () {
        patterns = new Pattern[core_patterns.length];
        pattern_buttons = new Gtk.RadioButton[patterns.length];
        pattern_box = new Gtk.VBox (true, 0);
        add_with_viewport (pattern_box);
        invisible_button = new Gtk.RadioButton (null);
        for (int i=0; i < patterns.length; i++) {
            patterns[i] = core_patterns[i];
            pattern_box.add (pattern_buttons[i] = make_button (i));
        }
        set_policy (Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS);
    }

    /**
     * Add custom patterns, replacing any previous ones.
     */
    public void add_patterns (Pattern[] additional_patterns) {
        // Remove previous ones
        for (int i = core_patterns.length; i < patterns.length; i++) {
            patterns[i] = null;
            pattern_buttons[i].destroy ();
            pattern_buttons[i] = null;
        }
        patterns.length = core_patterns.length;
        pattern_buttons.length = core_patterns.length;
        
        // Add new ones
        foreach (var p in additional_patterns) {
            patterns += p;
            var button = make_button (patterns.length - 1);
            pattern_buttons += button;
            pattern_box.pack_start (button, false, false, 0);
        }
        pattern_box.show_all ();
    }

    /**
     * Create a radio button to select a pattern (given by its index).
     */
    private Gtk.RadioButton make_button (int index) {
        var button = new Gtk.RadioButton (invisible_button.get_group ());
        button.set_mode (false);  // On/off only. No intermediate state.
        button.border_width = 0;
        button.relief = Gtk.ReliefStyle.NONE;
        button.image = new Gtk.Image.from_pixbuf (patterns[index].pixbuf);
        // Use signal_clicked rather than signal_toggled to allow a pattern randomizer
        // to be clicked again (despite being already selected)
        button.clicked.connect ((s) => {on_selection (index);});
        string pattern_name = _(patterns[index].name);
        button.tooltip_text =
            (patterns[index] is RandomPattern) ? _("Random Pattern: %s").printf (pattern_name) :
            (patterns[index] is CustomPattern) ? _("Custom-made Pattern") :
                                                 _("Pattern: \"%s\"").printf (pattern_name);
        return button;
    }

    /**
     * Select a random pattern. This can be a traditional pattern, or a pattern
     * generated by a randomizer.
     */
    public void random () {
        int rnd = Random.int_range (0, patterns.length);
        unselect ();
        self_clicked = true;
        pattern_buttons[rnd].clicked ();
        self_clicked = false;
    }

    /**
     * Handler for button click. To "click" the same button again, call on_selection (-1).
     */
    public void on_selection (int pattern_num) {
        if (!self_clicked && (ignore_signal || !responsive)) return;
        
        // Test for exceptional case where a previously selected custom pattern was subsequently removed
        if (mru_pattern_num >= patterns.length) mru_pattern_num = patterns.length - 1;

        // mru_pattern_num can be -1 for a pattern that was sent by a remote opponent
        if (mru_pattern_num == -1) mru_pattern_num = 0;

        // Re-select the previously selected pattern
        if (pattern_num == -1) {
            pattern_num = mru_pattern_num;
            ignore_signal = true;
            pattern_buttons[pattern_num].active = true;
        }
        // Ignore de-selection event. (Is this ever triggered?)
        if (pattern_buttons[pattern_num].active == false) return;
        
        // Scroll the selection into view
        if (is_realized () && pattern_buttons[pattern_num].active) {
            double button_size = vadjustment.upper / patterns.length;
            vadjustment.clamp_page (button_size * pattern_num, button_size * (pattern_num + 1));
        }

        // Turn on the selected random-pattern button, and randomize if it is selected while "hot"
        if (!ignore_signal && (patterns[pattern_num] is RandomPattern)) {
            bool hot = pattern_num == mru_pattern_num && timer != null && timer.elapsed () < 9.0;
            if (timer == null) timer = new Timer ();
            timer.reset ();
            if (hot || self_clicked) {
                (patterns[pattern_num] as RandomPattern).next_random ();
            }
            else {
                Timeout.add (200, () => {
                    double temperature = 3.0 - double.min (double.max (timer.elapsed () / 3, 2.0), 3.0);
                    if (pattern_num != mru_pattern_num) temperature = 0;
                    (patterns[pattern_num] as RandomPattern).temperature = temperature;
                    pattern_buttons[pattern_num].queue_draw ();
                    return temperature != 0;
                });
            }
            (patterns[pattern_num] as RandomPattern).temperature = self_clicked ? 0.0 : 1.0;
            pattern_buttons[pattern_num].queue_draw ();
        }
        ignore_signal = false;
        mru_pattern_num = pattern_num;
        pattern_changed_signal (patterns[pattern_num].dots);
    }

    /**
     * Unselect all buttons by selecting the invisible one.
     */
    public void unselect () {
        ignore_signal = true;
        invisible_button.set_active (true);
        ignore_signal = false;
    }

    /**
     * Apply set_sensitive to all child, except the currently selected pattern.
     */
    public new void set_sensitive (bool b) {
        responsive = b;
        foreach (var button in pattern_buttons) button.set_sensitive (b);
        if (mru_pattern_num >= patterns.length) mru_pattern_num = patterns.length - 1;
        if (mru_pattern_num != -1) pattern_buttons[mru_pattern_num].set_sensitive (true);
    }

    /**
     * Find a pattern. Set the selection to this pattern.
     * Trigger pattern_changed_signal, even if the pattern is not found.
     */
    public void find_and_click (string dots) {
        mru_pattern_num = -1;
        var compressed_dots = dots.replace (" ", "");
        for (int i=0; i < patterns.length; i++) {
            if (!(patterns[i] is RandomPattern) && patterns[i].dots.replace (" ", "") == compressed_dots) {
                // Select the pattern
                ignore_signal = true;
                pattern_buttons[i].set_active (true);
                mru_pattern_num = i;
                ignore_signal = false;
                // Scroll the selection into view
                double button_size = vadjustment.upper / patterns.length;
                vadjustment.clamp_page (button_size * i, button_size * (i + 1));
                break;
            }
        }
        pattern_changed_signal (dots);
    }
}//class
}//namespace
// vim: tabstop=4: expandtab: textwidth=100: autoindent:
