/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

struct Rgb {
    public double red;
    public double green;
    public double blue;
    public Rgb (double r, double g, double b) {
        red = r;
        green = g;
        blue = b;
    }
}

class Theme2DCairo : Theme2D {
    private string path = null;
    private int prev_board_size = -1;
    private int num_frames = 40;

    public Theme2DCairo () {
        items += ThemeItem () {name=_("Spin"), path="0", engine=this};
        items += ThemeItem () {name=_("Dizzy"), path="dizzy", engine=this};
        items += ThemeItem () {name=_("Nothing Good On TV"), path="tv", engine=this};
        items += ThemeItem () {name=_("Smooth and Shiny"), path="shiny", engine=this};
        items += ThemeItem () {name=_("Enchanted Flowers"), path="flower", engine=this};
        items += ThemeItem () {name=_("Galaxy"), path="galaxy", engine=this};
    }

    public override Theme? select (string path, int width, int height) {
        time_span = 2.0;
        prev_board_size = -1;
        this.path = path;
        resize (width, height);
        return this;
    }

    public override void resize (int width, int height) {
        int board_size = int.min (width, height);
        if (board_size == prev_board_size) return;
        prev_board_size = board_size;
        int piece_size = 66 * board_size / 572;
        int kas_cycle = num_frames - 1;
        int piece_cycle = num_frames / 2 - 1;
        switch (path) {
        case "0" :
            cycles = {kas_cycle, kas_cycle, piece_cycle, piece_cycle};
            make_pieces (piece_size, yin_yang);
            make_board (board_size, 2, Rgb (0.2, 0.5, 0.1),  Rgb (1, 1, 0), 0.5);
            break;
        case "dizzy" :
            cycles = {kas_cycle, kas_cycle, piece_cycle, piece_cycle};
            make_pieces (piece_size, ring);
            make_board (board_size, 3, Rgb (0.2, 0.5, 0.7),  Rgb (1, 1, 1), 0.2);
            break;
        case "tv" :
            cycles = {- kas_cycle, - kas_cycle, num_frames / 4 - 1, num_frames / 4 - 1};
            make_pieces (piece_size, tv);
            make_board (board_size, 5, Rgb (0.7, 0.5, 0.2),  Rgb (0, 0, 0), 0.5);
            break;
        case "shiny" :
            cycles = {0, 0, 0, 0};
            make_pieces (piece_size, shiny);
            make_board (board_size, board_size * 2, Rgb (0.6, 0.4, 0.3), Rgb (1, 1, 1), 0.5);
            break;
        case "flower" :
            cycles = {piece_cycle, piece_cycle, 0, 0};
            make_pieces (piece_size, flower);
            make_board (board_size, 2, Rgb (0.35, 0.3, 0.2), Rgb (1, 1, 1), 0.5);
            break;
        case "galaxy" :
            kas_cycle = - num_frames * 3 / 4;
            cycles = {kas_cycle, kas_cycle, piece_cycle, - piece_cycle};
            make_pieces (piece_size, galaxy);
            galaxy_board (board_size);
            break;
        default :
            assert_not_reached ();
        }
    }

    private void make_pieces (int size, Smith smith) {
        for (int i=0; i < Piece.COUNT; i++) {
            var piece = (Piece) i;
            images[i] = new Cairo.ImageSurface (Cairo.Format.ARGB32, size * num_frames, size);
            var cr = new Cairo.Context (images[i]);
            for (int f=0; f < num_frames; f++) {
                cr.save ();
                cr.translate (size * f + size / 2, size / 2);
                smith (cr, piece, size, f);
                cr.restore ();
            }//for f
        }//for i
    }

    private void shine_light (Cairo.Context cr, double brightness, double radius, double tilt) {
        var spot_angle = 0.25 * Math.PI - 0.75 * Math.PI * tilt;
        var spot_radius = radius * (0.5 + 0.3 * tilt);
        var spot_x = spot_radius * Math.cos (spot_angle);
        var spot_y = spot_radius * Math.sin (spot_angle);
        var shade = new Cairo.Pattern.radial (-spot_x, - spot_y, 0, 0, 0, radius);
        shade.add_color_stop_rgba (0.0, 1, 1, 1, 0.95 * brightness);
        shade.add_color_stop_rgba (0.3, 1, 1, 1, 0.5 * brightness);
        shade.add_color_stop_rgba (1.0, 0.2, 0.2, 0.2, 0.2 * brightness);
        cr.set_source (shade);
        cr.arc (0, 0, radius, 0, 2 * Math.PI);
        cr.fill ();
    }

    /**
     * Create an image for frame f of the given piece.
     */
    delegate void Smith (Cairo.Context cr, Piece piece, int size, int f);

    private void flower (Cairo.Context cr, Piece piece, int size, int f) {
        cr.save ();
        var radius = 0.33 * size;
        var frac = 1.0 * f / (num_frames / 2);
        var frac2 = double.min (1.0, frac);
        var r_out = (1.0 + frac2) * radius;
        var r_in  = (1.0 - frac2) * radius;
        // Color
        double[] reds =   {0.8, 0.3, 0.2, 1.0};
        double[] greens = {0.2, 0.2, 0.2, 1.0};
        double[] blues =  {0.3, 0.8, 0.2, 1.0};
        var red = reds[piece];
        var green = greens[piece];
        var blue = blues[piece];
        //cr.set_source_rgb (red, green, blue);
        var color = new Cairo.Pattern.radial (0, 0, 0, 0, 0, (2.0 - frac2) * radius);
        color.add_color_stop_rgb (0, red, green, blue);
        color.add_color_stop_rgb (0.5, red, green, blue);
        color.add_color_stop_rgb (1.0, 1.0, 1.0, 0);
        cr.set_source (color);
        cr.move_to (r_in, 0);
        var num_corners = piece.is_kas () ? 16 : 8;
        for (int i=0; i <= num_corners; i++) {
            if (!piece.is_kas () && f > num_frames / 2 && (frac - 1) * (num_corners + 1) >= i) continue;
            var shift = piece.is_kas () && f > num_frames / 2 ? (frac - 1) * 2 * Math.PI / num_corners : 0.0;
            var angle = Math.PI * 2 * i / num_corners + shift;
            var x1 = r_out * Math.cos (angle);
            var y1 = r_out * Math.sin (angle);
            angle = Math.PI * 2 * (i + 1) / num_corners + shift;
            var x2 = r_out * Math.cos (angle);
            var y2 = r_out * Math.sin (angle);
            var x3 = r_in * Math.cos (angle);
            var y3 = r_in * Math.sin (angle);
            cr.curve_to (x1, y1, x2, y2, x3, y3);
        }
        cr.clip ();
        cr.paint ();

        // Roughen
        double[,] p  = { {0, 0, 1, 1}, {0, 1, 1, 0}, {0, 0, 1, 0}, {0, 0, 0, 1} };
        for (int d=0; d < 4; d++) {
            var sheet = 4096;
            var shade = new Cairo.Pattern.linear (p[d, 0] * sheet - size, 
                                                  p[d, 1] * sheet - size,
                                                  p[d, 2] * sheet - size,
                                                  p[d, 3] * sheet- size);
            int v = 0;
            for (double i=0; i <= size * 2; i += 1.5) {
                v++;
                double val = v % 2;
                shade.add_color_stop_rgba (i / sheet, val, val, val, 0.05);
            }
            cr.set_source (shade);
            cr.paint ();
        }
        cr.restore ();
    }

    private void shiny (Cairo.Context cr, Piece piece, int size, int f) {
        cr.save ();
        var radius = 0.30 * size;
        // Lift
        var n1 = int.min (num_frames - 2, piece.is_kas () ? f : f * 2);
        var n2 = num_frames;
        cr.translate (0, - n1 * size / 3.0 / n2);
        cr.scale (1.0, (double) (n2 - n1 / 2) / n2);
        // Color
        double[] reds =   {1.0, 0.3, 0.2, 1.0};
        double[] greens = {0.2, 0.2, 0.2, 1.0};
        double[] blues =  {0.3, 1.0, 0.2, 1.0};
        var red = reds[piece];
        var green = greens[piece];
        var blue = blues[piece];
        var color = new Cairo.Pattern.radial (0, 0, 0, 0, 0, radius);
        color.add_color_stop_rgba (0, red, green, blue, 0.98);
        color.add_color_stop_rgba (0.8, red, green, blue, 1.0);
        cr.set_source (color);
        cr.set_source_rgb (red, green, blue);
        cr.arc (0, 0, radius, 0, 2 * Math.PI);
        cr.fill ();
        shine_light (cr, 1.0, radius, 1.0 * n1 / n2);
        cr.restore ();
    }

    private void yin_yang (Cairo.Context cr, Piece piece, int size, int f) {
        cr.save ();
        var n_rotation = piece.is_kas () ? 1 : 2;
        cr.rotate (n_rotation * 2 * Math.PI * f / (num_frames - 1));
        Rgb[] rgb1 = {Rgb (1.0, 0.1, 0.3), Rgb (0.1, 0.3, 1.0), Rgb (0.0, 0.0, 0.0), Rgb (0.8, 0.8, 0.8)};
        Rgb[] rgb2 = {Rgb (1.0, 0.4, 0.6), Rgb (0.4, 0.6, 1.0), Rgb (0.2, 0.2, 0.2), Rgb (0.9, 0.9, 0.9)};
        Rgb dark = rgb1[piece];
        Rgb light = rgb2[piece];
        var radius = size * (piece.is_kas () || f < num_frames / 2 ? 0.3 : 0.6 - 0.6 * f / (num_frames - 1));
        var color1 = new Cairo.Pattern.radial (0, 0, 0, 0, 0, radius);
        color1.add_color_stop_rgba (0.0, dark.red, dark.green, dark.blue, 0.9);
        color1.add_color_stop_rgba (0.8, dark.red, dark.green, dark.blue, 0.95);
        color1.add_color_stop_rgba (1.0, dark.red, dark.green, dark.blue, 1.0);
        cr.set_source (color1);
        cr.arc (0, 0, radius, 0, Math.PI);
        cr.arc (-radius / 2, 0, radius / 2, Math.PI, 2 * Math.PI);
        cr.arc_negative (radius / 2, 0, radius / 2 - 0.5, Math.PI, 0);
        cr.fill ();
        var color2 = new Cairo.Pattern.radial (0, 0, 0, 0, 0, radius);
        color2.add_color_stop_rgba (0.0, light.red, light.green, light.blue, 0.9);
        color2.add_color_stop_rgba (0.8, light.red, light.green, light.blue, 0.95);
        color2.add_color_stop_rgba (1.0, light.red, light.green, light.blue, 1.0);
        cr.set_source (color2);
        cr.arc (0, 0, radius, Math.PI, 2 * Math.PI);
        cr.arc (radius / 2, 0, radius / 2, 0, Math.PI);
        cr.arc_negative (-radius / 2, 0, radius / 2 - 0.5, 2 * Math.PI, Math.PI);
        cr.fill ();
        cr.restore ();
        shine_light (cr, 1.0, radius, 0.0);
    }

    private void ring (Cairo.Context cr, Piece piece, int size, int f) {
        Rgb[] rgb = {Rgb (1.0, 0.0, 0.0), Rgb (0.0, 0.0, 1.0), Rgb (0.0, 0.0, 0.0), Rgb (1.0, 1.0, 1.0)};
        Rgb paint = rgb[piece];
        var phase = Math.remainder (4 * Math.PI * f / (num_frames - 1), 2 * Math.PI);
        var num_rings = 5;
        var band = 0.25;
        var radius = size * (piece.is_kas () ? 0.3 : 0.3 - 0.05 * Math.sin (phase));
        var alpha = 1.0;
        if (! piece.is_kas () && f > num_frames / 2) {
            // fade = 1 to 0
            var fade = (Math.cos (Math.PI * (f - num_frames / 2) / (num_frames / 2)) + 1) / 2;
            band += 0.9 * (1 - fade);
            alpha *= fade;
            radius *= (1 + (1 - fade) / 2);
        }
        var delta = phase / (2 * Math.PI);
        cr.arc (0, 0, radius, 0, 2 * Math.PI);
        for (int ring = num_rings; ring >= 0; ring--) {
            var r = (delta + ring) * radius / num_rings;
            var r2 = (delta + ring - band) * radius / num_rings;
            if (r > radius) continue;
            cr.arc_negative (0, 0, r, 2 * Math.PI, 0);
            cr.arc (0, 0, r2, 0, 2 * Math.PI);
        }
        var color = new Cairo.Pattern.radial (0, 0, 0, 0, 0, radius);
        color.add_color_stop_rgba (0.0, paint.red, paint.green, paint.blue, 1.0 * alpha);
        color.add_color_stop_rgba (0.8, paint.red, paint.green, paint.blue, 0.8 * alpha);
        color.add_color_stop_rgba (1.0, paint.red, paint.green, paint.blue, 0.0 * alpha);
        cr.set_source (color);
        cr.fill ();
        shine_light (cr, alpha, radius, 0.0);
    }

    private void tv (Cairo.Context cr, Piece piece, int size, int f) {
        if (piece.is_kas ())
            tv_control (cr, piece, size, f);
        else
            tv_box (cr, piece, size, f);
    }

    private void tv_control (Cairo.Context cr, Piece piece, int size, int f) {
        Rgb paint = piece == Piece.KAS_0 ? Rgb (1.0, 0.3, 0.3) : Rgb (0.3, 0.3, 1.0);
        var width = 0.3 * size;
        var height = 0.5 * size;
        var x = -0.15 * size;
        var y = -0.25 * size;
        cr.set_source (new Cairo.Pattern.rgb (paint.red, paint.green, paint.blue));
        cr.rectangle (x, y, width, height);
        cr.fill ();
        // Relative size of buttons, gaps and borders :
        //        __[ ]_[ ]_[ ]__ 
        // 0.3 + 1 + 0.1 + 1 + 0.1 + 1 + 0.3 = 3.8
        var proportion = 3.8;
        var button_width = width / proportion;
        int pressed_button = f == 0 ? -1 : Random.int_range (0, 15);
        for (int i=0; i < 15; i++) {
            var delta = i == pressed_button ? 0.15 * button_width : 0;
            var col = i % 3;
            var row = i / 3;
            var x2 = x + (0.3 + col * 1.1) / proportion * width;
            var y2 = y + (0.3 + row * 1.1) / proportion * width;
            cr.set_source (new Cairo.Pattern.rgba (0, 0, 0, 0.5));
            cr.rectangle (x2 + delta, y2 + delta, button_width - delta, button_width - delta);
            cr.fill ();
        }
    }

    private void tv_box (Cairo.Context cr, Piece piece, int size, int f) {
        Rgb paint = piece == Piece.BLACK ? Rgb (0.3, 0.3, 0.3) : Rgb (0.8, 0.8, 0.8);
        // Box
        var xy = -0.3 * size;
        var width = 0.6 * size;
        var color = new Cairo.Pattern.linear (xy, xy, xy, width);
        color.add_color_stop_rgb (0, paint.red, paint.green, paint.blue);
        color.add_color_stop_rgb (1, paint.red, paint.green, paint.blue);
        cr.rectangle (xy, xy, width, width);
        cr.set_source (color);
        cr.fill ();
        if (f == 0) return;
        // White screen
        xy = -0.25 * size;
        width = 0.5 * size;
        var exposed = double.min (f / (num_frames / 4.0), 1) * width / 2 - 1;
        cr.rectangle (-exposed, xy, exposed * 2, width);
        cr.set_source (new Cairo.Pattern.rgb (1, 1, 1));
        cr.fill ();
        // Text on screen
        if (f > num_frames / 2) {
            var text = piece == Piece.BLACK ? "1" : "2";
            Cairo.TextExtents te;
            cr.select_font_face ("Serif", Cairo.FontSlant.NORMAL, Cairo.FontWeight.BOLD);
            cr.set_font_size (0.8 * width);
            cr.text_extents (text, out te);
            cr.set_source (new Cairo.Pattern.rgb (0.0, 0.4, 0.2));
            cr.move_to (0.5 - te.width / 2 - te.x_bearing, 0.5 - te.height / 2 - te.y_bearing);
            cr.show_text (text);
        }
        // Noise
        for (double x= -exposed; x < exposed; x += 1)
            for (double y=1; y < width - 1; y += 1)
                if (f < num_frames / 2) {
                    if (Random.boolean ()) {
                        cr.set_source (new Cairo.Pattern.rgb (0, 0, 0));
                        cr.rectangle (x, xy + y, 1, 1);
                        cr.fill ();
                    }
                }
                else if (Random.boolean () && f < num_frames * 3 / 4) {
                    cr.set_source (Random.boolean () ? new Cairo.Pattern.rgb (0, 0, 0) : new Cairo.Pattern.rgb (1, 1, 1));
                    cr.rectangle (x, xy + y, 1, 1);
                    cr.fill ();
                }
        // Fuzzy line
        if (f < num_frames / 2) {
            var fuzzy_line = xy + 1 + f % (num_frames / 4) / (num_frames / 4.0) * (width - 3);
            cr.rectangle (-exposed, fuzzy_line, exposed * 2, 2);
            cr.fill ();
        }
    }

    private double[,] stars = null;

    private void galaxy (Cairo.Context cr, Piece piece, int size, int f) {
        cr.save ();
        cr.arc (0, 0, size / 2, 0, 2 * Math.PI);
        cr.clip ();
        switch (piece) {
        case Piece.BLACK :
            galaxy_black (cr, size, f);
            break;
        case Piece.WHITE :
            galaxy_white (cr, size, f);
            break;
        default :
            galaxy_kas (cr, piece, size, f);
            break;
        }
        cr.restore ();
    }

    private void galaxy_kas (Cairo.Context cr, Piece piece, int size, int f) {
        int n_corners = 20;
        if (f == 0) {
            stars = new double[n_corners * 2, 2];
            // Phase randomizer
            for (int i=0; i < n_corners * 2; i++) {
                stars[i, 0] = Random.next_double () * 2 * Math.PI;   // Theta
                stars[i, 1] = Random.next_double () * 2 * Math.PI;   // Length of control tangents
            }
        }
        var phase = 4 * Math.PI * f / num_frames;
        // 0 <= distort <= 1, starting from 0 at frame 0
        var distort = f > num_frames / 2 ? 1.0 : Math.sin (Math.PI / 2 * f / (num_frames / 2));
        var radius = (0.2 + 0.3 * distort) * size;
        var corona = 0.7 * radius;
        for (int layer=0; layer < corona; layer++) {
            radius -= 1.0;
            // End points
            double[] fx = new double[n_corners];
            double[] fy = new double[n_corners];
            // Control points, 2 for each end points.
            double[] cx = new double[2 * n_corners];
            double[] cy = new double[2 * n_corners];
            for (int i=0; i < n_corners; i++) {
                // Fixed points around a cirle
                var theta = 2 * Math.PI * (i + 1) / n_corners;
                // Rotate the circle
                //theta += 4 * Math.PI * f / num_frames;
                fx[i] = -radius * Math.cos (theta);
                fy[i] = -radius * Math.sin (theta);
                // Make theta perpendicular to circle
                theta += Math.PI / 2;
                // theta += angle between -pi/3 and pi/3
                theta += Math.sin (phase + stars[i, 0]) * Math.PI / 3 * distort;
                var n = i * 2;
                var len = (0.2 + 0.2 * Math.sin (distort * stars[n, 1] + phase)) * radius;
                cx[n] = fx[i] + len * Math.cos (theta);
                cy[n] = fy[i] + len * Math.sin (theta);
                len = (0.2 + 0.2 *  Math.sin (distort * stars[n + 1, 1] + phase)) * radius;
                cx[n + 1] = fx[i] - len * Math.cos (theta);
                cy[n + 1] = fy[i] - len * Math.sin (theta);
            }
            cr.move_to (fx[0], fy[0]);
            for (int i=0; i < n_corners; i++) {
                var j = (i + 1) % n_corners;   // Index to next end point
                cr.curve_to (cx[2 * i + 1], cy[2 * i + 1],  cx[2 * j] , cy[2 * j],  fx[j], fy[j]);
            }
            var alpha = Math.pow (layer / corona, 3) * 0.6;
            if (piece == Piece.KAS_0)
                cr.set_source (new Cairo.Pattern.rgba (1, alpha, alpha, alpha));
            else
                cr.set_source (new Cairo.Pattern.rgba (alpha, alpha, 1, alpha));
            cr.fill ();
        }//for layer
    }

    private void galaxy_black (Cairo.Context cr, int size, int f) {
        if (f == num_frames - 1) return;
        int num_stars = size * size / 20;
        if (f == 0) {
            stars = new double[num_stars, 5];
            for (int i=0; i < num_stars; i++) {
                // Box-Muller transform: uniformly to normally distributed random numbers
                var r = 0.04 * size * (num_stars - i / 2) / num_stars * Math.sqrt (- 2 * Math.log (Random.next_double ()));
                var theta = 2 * Math.PI * Random.next_double ();
                stars[i, 0] = r * Math.cos (theta);
                stars[i, 1] = r * Math.sin (theta);
                // Random colors
                stars[i, 2] = Random.next_double () * 0.7;
                stars[i, 3] = Random.next_double () * 0.7;
                stars[i, 4] = Random.next_double () * 0.7;
            }
        }
        for (int i=0; i < num_stars; i++) {
            var x = stars[i, 0];
            var y = stars[i, 1];
            // Polar equation
            var theta = 1.5 * Math.PI * i / num_stars - 4 * Math.PI * f / num_frames;
            var r =  0.2 * size * i / num_stars;
            if (f >= num_frames / 2) {
                var t = (1.0 * f - num_frames / 2) / (num_frames / 2);
                // r = pulsating star band          + expanding central hole
                r = (1 - (t - 0.3) * (t - 0.3)) * r + 0.55 * t * t * size;
                x *= 1 + t;
                y *= 1 + t;
            }
            var x0 = r * Math.cos (theta) + x;
            var y0 = r * Math.sin (theta) + y;
            var x1 = -r * Math.cos (theta) + y;
            var y1 = -r * Math.sin (theta) + x;
            cr.rectangle (x0, y0, 1, 1);
            cr.rectangle (x1, y1, 1, 1);
            cr.set_source (new Cairo.Pattern.rgb (stars[i, 2], stars[i, 3], stars[i, 4]));
            cr.fill ();
        }
    }

    private void galaxy_white (Cairo.Context cr, int size, int f) {
        if (f == num_frames - 1) return;
        int num_stars = size * size / 10;
        if (f == 0) {
            stars = new double[num_stars, 4];
            for (int i=0; i < num_stars; i++) {
                // Box-Muller transform: uniformly to normally distributed random numbers
                var r = 0.08 * size * Math.sqrt (- 2 * Math.log (Random.next_double ()));
                var theta = 2 * Math.PI * Random.next_double ();
                stars[i, 0] = r * Math.cos (theta);
                stars[i, 1] = r * Math.sin (theta);
                // Random phases
                stars[i, 2] = Random.int_range (0, num_frames);
                stars[i, 3] = Random.int_range (0, num_frames);
           }
        }
        var t = int.max (1, f - num_frames / 2 + 2);
        for (int i=0; i < num_stars; i++) {
            var x = stars[i, 0];
            var y = stars[i, 1];
            if (t > 1) {
                // Explode the galaxy.
                // Expansion rate grows with time. Repulsion force is greatest in the centre.
                var r = Math.hypot (x, y);
                var force = (size - r) / size;
                var expand = 1.0 + (0.005 * t * t) * (force * force);
                stars[i, 0] = (x *= expand);
                stars[i, 1] = (y *= expand);
            }
            if (i >= 0.95 * num_stars) {
                // Twinkling stars
                var phase0 = stars[i, 2] + f;
                var phase1 = stars[i, 3] + f;
                var light0 = 0.75 + 0.25 * Math.sin (2 * Math.PI * phase0 / num_frames);
                var light1 = 0.75 + 0.25 * Math.sin (2 * Math.PI * phase1 / num_frames);
                cr.arc (x, y, light0 * 2, 0, 2 * Math.PI);
                cr.set_source (new Cairo.Pattern.rgb (light0, light0, light1));
            }
            else {
                // Static background stars
                cr.rectangle (x, y, 1, 1);
                cr.set_source (new Cairo.Pattern.rgb (0.7, 0.7, 0.7));
            }
            cr.fill ();
        }
    }

    private void galaxy_board (int size) {
        images[Piece.BOARD] = new Cairo.ImageSurface (Cairo.Format.ARGB32, size, size);
        var cr = new Cairo.Context (images[Piece.BOARD]);
        var shade = new Cairo.Pattern.linear (size / 3, size, size, 0);
        shade.add_color_stop_rgb (0, 0.1, 0.2, 0.0);
        shade.add_color_stop_rgb (1, 0.0, 0.1, 0.3);
        cr.set_source (shade);
        cr.paint ();

        var gap = cell_width ();
        for (var y = 1.5 * gap; y <= 11.5 * gap; y += gap) {
            for (var x = 1.5 * gap; x <= 11.5 * gap; x += 2) {
                if (y > 5.5 * gap && y < 7.5 * gap && x > 5.5 * gap && x < 7.5 * gap) continue;
                var r0 = 1 - 2 * Random.next_double ();
                var r1 = 1 - 2 * Random.next_double ();
                var z0 = r0 * r0 * r0 * gap * 0.8;
                var z1 = r1 * r1 * r1 * gap * 0.8;
                var star = 0.4 + 0.6 * (1 - Math.fabs (r0));
                cr.set_source_rgb (star, star, star);
                cr.rectangle (x, y + z0, 1, 1);
                cr.rectangle (y + z1, x, 1, 1);
                cr.fill ();
            }
        }
    }
        
    private void make_board (int size, int grain_size, Rgb back, Rgb line, double line_alpha) {
        images[Piece.BOARD] = new Cairo.ImageSurface (Cairo.Format.ARGB32, size, size);
        var tile_size = int.min (grain_size * 16, size);
        var tile = new Cairo.ImageSurface (Cairo.Format.ARGB32, tile_size, tile_size);
        var cr = new Cairo.Context (tile);
        cr.set_source_rgb (back.red, back.green, back.blue);
        cr.paint ();

        // Rough texture
        double[,] p  = { {0, 0, 1, 1}, {0, 1, 1, 0}, {0, 0, 1, 0}, {1, 0, 0, 1} };
        for (int d=0; d < 4; d++) {
            var sheet = 4096;
            var shade = new Cairo.Pattern.linear (p[d, 0] * sheet, p[d, 1] * sheet, p[d, 2] * sheet, p[d, 3] * sheet);
            int v = 0;
            for (int i=0; i <= tile_size * 2; i += grain_size) {
                v++;
                double val = v % 2;
                shade.add_color_stop_rgba ((double) i / sheet, val, val, val, 0.05);
            }
            cr.set_source (shade);
            cr.paint ();
        }
        // images[Piece.board] = tile repeated
        cr = new Cairo.Context (images[Piece.BOARD]);
        cr.set_source_surface (tile, 0, 0);
        cr.get_source().set_extend (Cairo.Extend.REPEAT);
        cr.paint ();

        // Horizontal and vertical lines
        cr.set_source_rgba (line.red, line.green, line.blue, line_alpha);
        int gap = cell_width ();
        for (int i=0; i < 11; i++) {
            var x = gap * 1.5 + gap * i - 0.03 * gap;
            var y = gap * 1.5 - 0.03 * gap;
            var w = 0.06 * gap;
            var h = 10 * gap;
            if (i != 5) {
                cr.rectangle (x, y, w, h + w);
                cr.rectangle (y, x, h, w);
            }
            else {
                cr.rectangle (x, y, w, 4 * gap);
                cr.rectangle (x, y + 6 * gap, w, 4 * gap);
                cr.rectangle (y, x, 4 * gap, w);
                cr.rectangle (y + 6 * gap, x, 4 * gap, w);
            }
        }
        cr.fill ();
    }

}//class
}//namespace
// vim: tabstop=4: expandtab: textwidth=100: autoindent:
