/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

abstract class Theme2D : Theme {
    protected Cairo.ImageSurface[] images = new Cairo.ImageSurface[Piece.COUNT + 1];

    public override void free () {
        for (int i=0; i < images.length; i++) images[i] = null;
    }

    public override void draw (GameView game_view, Gdk.Rectangle area) {
        var cr = Gdk.cairo_create (game_view.window);
        cr.rectangle (area.x, area.y, area.width, area.height);
        cr.clip ();

        Game game = game_view.game;
        Stage stage = game_view.game.stage;
        Move move = game_view.selected_move;
        Point kas_point = game_view.kas_point;
        bool rotated = game_view.rotated;
        draw_board (cr, 0, 0);
//        draw_score (cr, rotated, game);

        // TODO debug pieces
//        var pp = Piece.BLACK;
//        var xx = count_frames (pp);
//        for (int i=0; i <= xx/2; i++) {
//            draw_piece (cr, pp, i + xx / 2, piece_image_width (pp) * ((i) % 8), 
//                        i < 8 ? 0 : board_image_width () - piece_image_width (pp));
//        }

        for (int layer=0; layer < 150; layer++) {
            for (int position=0; position < BOARD_SIZE; position++) {
                if (films[position].viewed != layer) continue;

                // Skip blanks and guards
                Piece piece = game.board[position];
                if (! piece.is_real ()) continue;
                
                // If we are making a kas, the source square should be shown empty
                // (actually a piece is still there on the board because the move
                // is not completed/submitted yet).
                if (stage == Stage.SELECT && move != null && position == move.position) continue;

                // If we are dragging a kas, the source square should be empty.
                // (actually the kas is still there on the board, but we are going to
                // show the kas being dragged after this loop).
                if (stage == Stage.MOVE && kas_point.x != -1 && piece == game.kas) continue;

                Point p = to_point (position, rotated);
                int radius = piece_image_width (piece) / 2;
                draw_piece (cr, piece, films[position].viewed, p.x - radius,  p.y - radius);
            }//endfor position
            // Draw dragged kas
            if (films[game.kas].viewed != layer) continue;
            if ((stage == Stage.SELECT && move != null) || 
                (stage == Stage.MOVE && kas_point.x != -1)) {
                Piece piece = game.kas;
                int radius = piece_image_width (piece) / 2;
                draw_piece (cr, piece, films[game.kas].viewed, kas_point.x - radius,  kas_point.y - radius);
            }
        }//endfor layer
    }

    protected void draw_board (Cairo.Context cr, int x, int y) {
        cr.save ();
        int width = board_image_width ();
        cr.set_source_surface (images[Piece.BOARD], 0, 0);
        cr.rectangle (x, y, width, width);
        cr.fill ();
        cr.restore ();
    }

    public override void draw_piece (Cairo.Context cr, Piece piece, int frame, int x, int y)
        requires (piece.is_real ()) {
        cr.save ();
        frame = int.min (frame, count_frames (piece) - 1);
        int width = piece_image_width (piece);
        cr.translate (x, y);
        cr.set_source_surface (images[piece], - frame * width, 0);
        cr.rectangle (0, 0, width, width);
        cr.fill ();
        cr.restore ();
    }

//    public void draw_score (Cairo.Context cr, bool rotated, Game game) {
//        var width = board_image_width ();
//        var size = width / 60.0;
//        var x = width - size;
//        var y = cell_width () * 1.5;
//        var length = cell_width () * 10;
//        var ten = length / 18.0;
//        // Empty portion of the score bar
//        cr.save ();
//        var len = ten * (180 - game.score[0] - game.score[1]) / 10;
//        var offset = ten * (rotated ? game.score[1] : game.score[0]) / 10;
//        cr.rectangle (x, y + offset, size, len);
//        cr.clip ();
//        for (int i=0; i < 18; i++) {
//            var c = i % 2 == 0 ? 0.6 : 0.8;
//            cr.set_source_rgba (c, c, c, 0.6);
//            cr.rectangle (x, y + i * ten, size, ten);
//            cr.fill ();
//        }
//        cr.restore ();
//        // Red and blue portions of the score bar
//        for (int p=0; p < 2; p++) {
//            var player = rotated ? 1 - p : p;
//            cr.save ();
//            len = ten * game.score[player] / 10;
//            if (p == 0) cr.rectangle (x, y, size, len);
//            else        cr.rectangle (x, y + length - len, size, len);
//            cr.clip ();
//            for (int i=0; i < 18; i++) {
//                var alpha = 0.99;
//                var c = i % 2 == 0 ? 0.7 : 0.9;
//                if (player == 0) cr.set_source_rgba (c, 0.4 * c, 0.4 * c, alpha);
//                else             cr.set_source_rgba (0.4 * c, 0.4 * c, c, alpha);
//                cr.rectangle (x, y + i * ten, size, ten);
//                cr.fill ();
//            }
//            cr.restore ();
//        }
//    }

    public override bool detect_piece (Piece piece, int frame, int x, int y)
        requires (piece.is_real ()) {
        int width = piece_image_width (piece);
        x += width / 2;
        y += width / 2;
        frame = int.min (frame, count_frames (piece) - 1);
        var rgba = (uint32*)(images[piece].get_data () + 4 * (y * images[piece].get_width () + frame * width + x));
        return *rgba != 0;
    }

    public override int count_frames (Piece piece) 
        requires (piece.is_real ()) {
        return images[piece].get_width () / images[piece].get_height ();
    }

    public override int board_image_width () {
        return images[Piece.BOARD].get_width ();
    }

    public override int piece_image_width (Piece piece)
        requires (piece.is_real ()) {
        return images[piece].get_height ();
    }

    public override int cell_width () {
        return board_image_width () / BOARD_WIDTH;
    }

    public override Point to_point (int pos, bool rotated)
        requires (pos >= 0 && pos < BOARD_SIZE) {
        if (rotated) pos = BOARD_SIZE -1 - pos;
        int border_width = cell_width () / 2;
        return Point.xy (
            pos % BOARD_WIDTH * cell_width () + border_width,
            pos / BOARD_WIDTH * cell_width () + border_width
        );
    }

    public override int to_position (Game game, Point p, bool rotated, bool precise) {
        // Locate the nearest piece
        int col = p.x / cell_width ();
        int row = p.y / cell_width ();
        if (col < 1 || col > 11 || row < 1 || row > 11) return -1;
        int pos = row * BOARD_WIDTH + col;
        if (rotated) {
            pos = BOARD_SIZE - 1 - pos;
        }
        // Get the centre of this piece
        Point c = to_point (pos, rotated);
        // Relative position of p to c
        int x = p.x - c.x;
        int y = p.y - c.y;
        Piece piece = game.board[pos];
        int snap_range = 1 + (precise ? 1 : cell_width () / 6);  
        if // p is near c, even if the piece has a hole in the centre.
           ((p.x - c.x) * (p.x - c.x) + (p.y - c.y)*(p.y - c.y) <= snap_range * snap_range ||
           // p is touching a piece.
           !precise && piece.is_real () && detect_piece (piece, 0, x, y) ||
           // p is near a vacant position and the dragged kas is touching c.
           !precise && piece == Piece.EMPTY &&
            (detect_piece (game.kas, 0, -x, -y) || detect_piece (game.kas, films[0].viewed, -x, -y))) {
            return pos;
        }
        else {
            return -1;
        }
    }
}//class
}//namespace
// vim: tabstop=4: expandtab: textwidth=100: autoindent:
